/*
Converts UYVY to bitmap (.bmp or .tga file)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	TRUE		1
#define	FALSE		0

#define OPTION_CHAR	'-'

/* private prototypes */

void conv422to444(unsigned char *src, unsigned char *dst, int Coded_Picture_Width, int Coded_Picture_Height, int MPEG2_flag);

unsigned char *Clip;

/* color space conversion coefficients
 * for YCbCr -> RGB mapping
 *
 * entries are {crv,cbu,cgu,cgv}
 *
 * crv=(255/224)*65536*(1-cr)/0.5
 * cbu=(255/224)*65536*(1-cb)/0.5
 * cgu=(255/224)*65536*(cb/cg)*(1-cb)/0.5
 * cgv=(255/224)*65536*(cr/cg)*(1-cr)/0.5
 *
 * where Y=cr*R+cg*G+cb*B (cr+cg+cb=1)
 */

/* ISO/IEC 13818-2 section 6.3.6 sequence_display_extension() */

int Inverse_Table_6_9[8][4] =
{
  {117504, 138453, 13954, 34903}, /* no sequence_display_extension */
  {117489, 138438, 13975, 34925}, /* ITU-R Rec. 709 (2002) */
  {104597, 132201, 25675, 53279}, /* unspecified */
  {104597, 132201, 25675, 53279}, /* reserved */
  {104448, 132798, 24759, 53109}, /* FCC */
  {104597, 132201, 25675, 53279}, /* ITU-R Rec. 624-4 System B, G */
  {104597, 132201, 25675, 53279}, /* SMPTE 170M */
  {117579, 136230, 16907, 35559}  /* SMPTE 240M (1987) */
};

int Inverse_Table_6_9_Video[8][4] =
{
  {100915, 118907, 11984, 29976}, /* no sequence_display_extension */
  {100902, 118894, 12002, 29994}, /* ITU-R Rec. 709 (2002) */
  {89850, 113508, 22020, 45744},  /* unspecified */
  {89850, 113508, 22020, 45744},  /* reserved */
  {89702, 114050, 21264, 45611},  /* FCC */
  {89850, 113508, 22020, 45744},  /* ITU-R Rec. 624-4 System B, G */
  {89850, 113508, 22020, 45744},  /* SMPTE 170M */
  {100979, 116998, 14520, 30539}  /* SMPTE 240M (1987) */
};

int main(int argc, char **argv)
{
	FILE	*fp;
	FILE	*fpout;
	int	length, i, j, cr, cb, horz, vert, index, first_pixel = 0;
	int	y, u, v;
	unsigned	char	r, g, b;
	int	crv, cbu, cgu, cgv;
	int	tgaflag = TRUE;
	int	video_levels = FALSE;
	int	matrix_coefficients = 1;
	unsigned char	*py, *pu, *pv;
	unsigned char	horz1, horz2, vert1, vert2, temp;
	unsigned int	size;
	static unsigned char	tga24[14] = {0,0,2,0,0,0,0, 0,0,0,0,0,24,32};
	static unsigned char	bmp[54] = {0x42,0x4d,0,0,0,0,0,0,0,0,0x36,0,0,0,0x28,0,0,0,0,0,0,0,0,0,0,0,1,0,24,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	static unsigned char	buffer_422input[3840 * 2160 * 2];
	static unsigned char	buffer_422[3840 * 2160 * 2];
	static unsigned char	buffer_444Cb[3840 * 2160];
	static unsigned char	buffer_444Cr[3840 * 2160];
	static unsigned char	buffer_444output[3840 * 2160 * 3];

	if (argc != 5 && argc != 6) {
		fprintf(stderr, "usage: yuvtorgb <-bv6> <horz> <vert> <infile> <outfile>\n");
		exit(-1);
	}

	if(argc == 5)  {
		/*--- open binary file (for parsing) ---*/
		fp = fopen(argv[3], "rb");
		if (fp == 0) {
			fprintf(stderr, "Cannot open input file <%s>\n", argv[3]);
			exit(-1);
		}

		/*--- open binary file (for parsing) ---*/
		fpout = fopen(argv[4], "wb");
		if (fpout == 0) {
			fprintf(stderr, "Cannot open output file <%s>\n", argv[4]);
			exit(-1);
		}
		horz = atoi(argv[1]);
		vert = atoi(argv[2]);
	}
	else  {
		if(*argv[1] == OPTION_CHAR)  {
			for(i = 1; i < strlen(argv[1]); i++)  {
				switch (argv[1][i])  {
					case 'b':
					case 'B':
						tgaflag = FALSE;
						break;
					case 'v':
					case 'V':
						video_levels = TRUE;
						break;
					case '6':
						matrix_coefficients = 6;
						break;
					default:
						fprintf(stderr, "Unsupported Option: %c\n", argv[1][i]);
				}
			}
		}
		else  {
			fprintf(stderr, "usage: yuvtorgb <-bv6> <horz> <vert> <infile> <outfile>\n");
			exit(-1);
		}
		/*--- open binary file (for parsing) ---*/
		fp = fopen(argv[4], "rb");
		if (fp == 0) {
			fprintf(stderr, "Cannot open input file <%s>\n", argv[4]);
			exit(-1);
		}

		/*--- open binary file (for parsing) ---*/
		fpout = fopen(argv[5], "wb");
		if (fpout == 0) {
			fprintf(stderr, "Cannot open output file <%s>\n", argv[5]);
			exit(-1);
		}
		horz = atoi(argv[2]);
		vert = atoi(argv[3]);
	}

	/* Clip table */
	if (!(Clip=(unsigned char *)malloc(1024)))
		printf("Clip[] malloc failed\n");

	Clip += 384;

	for (i = -384; i < 640; i++)
		Clip[i] = (i<0) ? 0 : ((i>255) ? 255 : i);

	horz1 = horz;
	horz2 = horz >> 8;
	vert1 = vert;
	vert2 = vert>> 8;
	length = fread(&buffer_422input[0], 1, (horz * vert * 2), fp);
	y = 0;
	cb = horz * vert;
	cr = (horz * vert) + ((horz / 2) * (vert));
	for (i = 0; i < (horz * vert * 2); i++)
	{
		if ((i % 2) == 1)
		{
			buffer_422[y++] = buffer_422input[i];
		}
		else if ((i % 4) == 0)
		{
			buffer_422[cb++] = buffer_422input[i];
		}
		else
		{
			buffer_422[cr++] = buffer_422input[i];
		}
	}
	cb = horz * vert;
	cr = (horz * vert) + ((horz / 2) * (vert));
	conv422to444(&buffer_422[cb], &buffer_444Cb[0], horz, vert, 1);
	conv422to444(&buffer_422[cr], &buffer_444Cr[0], horz, vert, 1);

	if (tgaflag)
	{
		/* TGA header */
		for (i=0; i<12; i++)
			fwrite(&tga24[i], 1, 1, fpout);

		fwrite(&horz1, 1, 1, fpout);
		fwrite(&horz2, 1, 1, fpout);
		fwrite(&vert1, 1, 1, fpout);
		fwrite(&vert2, 1, 1, fpout);

		fwrite(&tga24[12], 1, 1, fpout);
		fwrite(&tga24[13], 1, 1, fpout);
	}
	else
	{
		/* BMP header */
		size = (horz * vert * 3) + 56;
		temp = (size >> 24) & 0xff;
		bmp[5] = temp;
		temp = (size >> 16) & 0xff;
		bmp[4] = temp;
		temp = (size >> 8) & 0xff;
		bmp[3] = temp;
		temp = size & 0xff;
		bmp[2] = temp;
		bmp[18] = horz1;
		bmp[19] = horz2;
		bmp[22] = vert1;
		bmp[23] = vert2;
		for (i=0; i<54; i++)
			fwrite(&bmp[i], 1, 1, fpout);
	}

	if (video_levels)
	{
		/* matrix coefficients */
		crv = Inverse_Table_6_9_Video[matrix_coefficients][0];
		cbu = Inverse_Table_6_9_Video[matrix_coefficients][1];
		cgu = Inverse_Table_6_9_Video[matrix_coefficients][2];
		cgv = Inverse_Table_6_9_Video[matrix_coefficients][3];
	}
	else
	{
		/* matrix coefficients */
		crv = Inverse_Table_6_9[matrix_coefficients][0];
		cbu = Inverse_Table_6_9[matrix_coefficients][1];
		cgu = Inverse_Table_6_9[matrix_coefficients][2];
		cgv = Inverse_Table_6_9[matrix_coefficients][3];
	}
  
	py = &buffer_422[0];
	pu = &buffer_444Cb[0];
	pv = &buffer_444Cr[0];

	if (tgaflag)
	{
		index = 0;
	}
	else
	{
		index = ((horz * (vert - 1)) * 3);
	}
	for (i=0; i<vert; i++)
	{
		for (j=0; j<horz; j++)
		{
			u = *pu++ - 128;
			v = *pv++ - 128;
			if (video_levels)
			{
				y = 65536 * (*py++);
			}
			else
			{
				y = 76309 * (*py++ - 16); /* (255/219)*65536 */
			}
			r = Clip[(y + crv*v + 32768)>>16];
			g = Clip[(y - cgu*u - cgv*v + 32768)>>16];
			b = Clip[(y + cbu*u + 32786)>>16];

			buffer_444output[index++] = b;
			buffer_444output[index++] = g;
			buffer_444output[index++] = r;
		}
		if (!tgaflag)
		{
			index = index - (horz * 6);
		}

	}
	fwrite(&buffer_444output[0], 1, (horz * vert * 3), fpout);
	fclose(fp);
	fclose(fpout);
	return 0;
}

/* horizontal 1:2 interpolation filter */
void conv422to444(unsigned char *src, unsigned char *dst, int Coded_Picture_Width, int Coded_Picture_Height, int MPEG2_flag)
{
  int i, i2, w, j, im3, im2, im1, ip1, ip2, ip3;

  w = Coded_Picture_Width>>1;

  if (MPEG2_flag)
  {
    for (j=0; j<Coded_Picture_Height; j++)
    {
      for (i=0; i<w; i++)
      {
        i2 = i<<1;
        im2 = (i<2) ? 0 : i-2;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<w-1) ? i+1 : w-1;
        ip2 = (i<w-2) ? i+2 : w-1;
        ip3 = (i<w-3) ? i+3 : w-1;

        /* FIR filter coefficients (*256): 21 0 -52 0 159 256 159 0 -52 0 21 */
        /* even samples (0 0 256 0 0) */
        dst[i2] = src[i];

        /* odd samples (21 -52 159 159 -52 21) */
        dst[i2+1] = Clip[(int)(21*(src[im2]+src[ip3])
                        -52*(src[im1]+src[ip2]) 
                       +159*(src[i]+src[ip1])+128)>>8];
      }
      src+= w;
      dst+= Coded_Picture_Width;
    }
  }
  else
  {
    for (j=0; j<Coded_Picture_Height; j++)
    {
      for (i=0; i<w; i++)
      {

        i2 = i<<1;
        im3 = (i<3) ? 0 : i-3;
        im2 = (i<2) ? 0 : i-2;
        im1 = (i<1) ? 0 : i-1;
        ip1 = (i<w-1) ? i+1 : w-1;
        ip2 = (i<w-2) ? i+2 : w-1;
        ip3 = (i<w-3) ? i+3 : w-1;

        /* FIR filter coefficients (*256): 5 -21 70 228 -37 11 */
        dst[i2] =   Clip[(int)(  5*src[im3]
                         -21*src[im2]
                         +70*src[im1]
                        +228*src[i]
                         -37*src[ip1]
                         +11*src[ip2]+128)>>8];

       dst[i2+1] = Clip[(int)(  5*src[ip3]
                         -21*src[ip2]
                         +70*src[ip1]
                        +228*src[i]
                         -37*src[im1]
                         +11*src[im2]+128)>>8];
      }
      src+= w;
      dst+= Coded_Picture_Width;
    }
  }
}
